"""liken main public API"""

from __future__ import annotations

import pandas as pd
import polars as pl
import pyspark.sql as spark
from pyspark.sql import SparkSession

from liken._dataframe import Frame
from liken._dataframe import wrap
from liken._executors import Executor
from liken._executors import LocalExecutor
from liken._executors import SparkExecutor
from liken._strats_library import BaseStrategy
from liken._strats_library import exact
from liken._strats_manager import Rules
from liken._strats_manager import StrategyManager
from liken._strats_manager import StratsDict
from liken._types import Columns
from liken._types import DataFrameLike
from liken._types import Keep
from liken._types import UserDataFrame
from liken._validators import validate_columns_arg
from liken._validators import validate_df_arg
from liken._validators import validate_keep_arg
from liken._validators import validate_spark_arg


# API:


class Dedupe:
    """Deduplicate a dataframe given a collection of strategies.

    Args:
        df: The dataframe to deduplicate.
        spark_session: optional spark session if initializing with PySpark
            backend.

    Attributes:
        df: Returns the dataframe as currently stored in the `Dedupe` instance's
            strategy manager.
        strats: Returns the strategies as currently stored in the strategy
            manager.

    Raises:
        ValueError: Initialized with PySpark DataFrame but no Spark Session.
    """

    _executor: Executor

    def __init__(
        self,
        df: UserDataFrame,
        /,
        *,
        spark_session: SparkSession | None = None,
    ):
        self._df: DataFrameLike = validate_df_arg(df)

        self._sm = StrategyManager()

        if isinstance(df, spark.DataFrame):
            spark_session = validate_spark_arg(spark_session)
            self._executor = SparkExecutor(spark_session=spark_session)
        else:
            self._executor = LocalExecutor()

    @classmethod
    def _from_rows(
        cls,
        rows: list[spark.Row],
    ) -> Dedupe:
        """bypass initialization and initialize explicitely with no validation.

        Use as internal constructor with spark `Rows`.
        """
        self = cls.__new__(cls)
        self._df = rows
        self._sm = StrategyManager()
        self._executor = LocalExecutor()
        return self

    def apply(self, strategy: BaseStrategy | dict | Rules) -> None:
        """Apply a strategy or strategies for deduplication.

        Available for inspection when access with attribute `.strats`. Can be
        repetitively called if using the Sequential API. Else apply once using
        the Dict API or Rules API.

        Args:
            strategy: The strategy or strategies to apply

        Returns:
            None

        Raises:
            InvalidStrategyError: For any invalid strategy or collection of
                strategies

        Example:
            Import and prepate data:

                from liken import Dedupe, exact, tfidf, lsh

                df = ... # get a dataframe

            Sequential API:

                lk = Dedupe(df)
                lk.apply(exact())
                lk.apply(tfidf()) # Ok, apply more than once
                df = lk.drop_duplicates("address")

            Dict API:

                lk = Dedupe(df)
                lk.apply(
                    {
                        "address": (exact(), tfidf()),
                    }
                )
                df = lk.drop_duplicates()

            Rules API:

                from liken.rules import Rules, on

                lk = Dedupe(df)
                lk.apply(
                    Rules(
                        on("address", exact()),
                        on("address", tfidf()),
                    )
                )
                df = lk.drop_duplicates()

        """
        self._sm.apply(strategy)

    def drop_duplicates(
        self,
        columns: Columns | None = None,
        *,
        keep: Keep = "first",
    ) -> pd.DataFrame | pl.DataFrame | spark.DataFrame:
        """Drop duplicates by enacting the applied strategies.

        If no strategies are explicitely provided, will carry out an exact
        deduplication on any number of columns provided in `columns`.
        The `.strats` attribute will return None after calling this function.

        Args:
            columns (str | tuple[str, ...] | None): The attribute(s) of the
                dataframe to deduplicate.
            keep: Accepted as "first" or "last". Whether to keep the first intance
                of a duplicate or the last intance, as found in the DataFrame.

        Returns:
            A deduplicated DataFrame.

        Raises:
            ValueError: Incorrect value to `keep` arg.
            ValueError: Incorrect use of `columns` arg given API used to apply
                strategies.
            ValueError: Incorrect use a single column strategy given multiple
                columns defined, or vice-versa.
        """
        keep: Keep = validate_keep_arg(keep)
        columns: Columns | None = validate_columns_arg(columns, self._sm.is_sequential_applied)
        wdf: Frame = wrap(self._df, None)  # canonical id only ever autoincremental for dropping

        # No .apply(), assumes exact deduplication
        if not self._sm.has_applies:
            self._sm.apply(exact())
        strats: StratsDict | Rules = self._sm.get()

        df = self._executor.execute(
            wdf,
            columns=columns,
            strats=strats,
            keep=keep,
            drop_duplicates=True,
            drop_canonical_id=True,
            id=None,
        )

        self._sm.reset()

        return df.unwrap()

    def canonicalize(
        self,
        columns: Columns | None = None,
        *,
        keep: Keep = "first",
        drop_duplicates: bool = False,
        id: str | None = None,
    ) -> pd.DataFrame | pl.DataFrame | spark.DataFrame:
        """Canonicalize by enacting the applied strategies.

        If no strategies are explicitely provided, will carry out an exact
        canonicalization on any number of columns provided in `columns`.
        The `.strats` attribute will return None after calling this function.

        Args:
            columns (str | tuple[str, ...] | None): The attribute(s) of the
                dataframe to deduplicate.
            keep: Accepted as "first" or "last". Whether to keep the first
                intance of a duplicate or the last intance, as found in the
                DataFrame.
            drop_duplicates: Optionally drop duplicates, whilst preserving a
                canonical_id, contrary to `drop_duplicates`.
            id: string label identifying a column in the dataframe that can be
                used to optionally override the values of a default
                canonical_id.

        Returns:
            A canonicalised DataFrame. By default canonicalization is tracked
                in a new `canonical_id` field.

        Raises:
            ValueError: Incorrect value to `keep` arg.
            ValueError: Incorrect use of `columns` arg given API used to apply
                strategies.
            ValueError: Incorrect use a single column strategy given multiple
                columns defined, or vice-versa.
        """
        keep: Keep = validate_keep_arg(keep)
        columns: Columns | None = validate_columns_arg(columns, self._sm.is_sequential_applied)
        wdf: Frame = wrap(self._df, id)

        # No .apply(), assumes exact deduplication
        if not self._sm.has_applies:
            self.apply(exact())
        strats: StratsDict | Rules = self._sm.get()

        df: Frame = self._executor.execute(
            wdf,
            columns=columns,
            strats=strats,
            keep=keep,
            drop_duplicates=drop_duplicates,
            drop_canonical_id=False,
            id=id,
        )

        self._sm.reset()

        return df.unwrap()

    @property
    def strats(self) -> str | None:
        """
        Returns the strategies as currently stored in the strategy manager.

        If no strategies are stored, returns None. Otherwise, returns a string
        representation of the strategies containedtuple
        of strategy names or a dictionary mapping attributes to their
        respective strategies.

        Returns:
            The stored strategies, formatted
        """
        return self._sm.pretty_get()



"""This moduel contains argument validation for classes.

Most validations are for public arguments of the 'Dedupe' class.

However, some validations exist for other private classes
"""

from collections import Counter
from typing import Literal

from pyspark.sql import SparkSession

from liken._constants import INVALID_COLUMNS_EMPTY
from liken._constants import INVALID_COLUMNS_NOT_NONE
from liken._constants import INVALID_COLUMNS_REPEATED
from liken._constants import INVALID_DF
from liken._constants import INVALID_KEEP
from liken._constants import INVALID_SPARK
from liken._constants import INVALID_STRAT
from liken._constants import SUPPORTED_DFS
from liken._strats_library import BaseStrategy
from liken._types import Columns
from liken._types import UserDataFrame


def validate_df_arg(df: UserDataFrame) -> UserDataFrame:
    if not isinstance(df, SUPPORTED_DFS):
        raise ValueError(INVALID_DF.format(type(df).__name__))
    return df


def validate_spark_arg(spark_session: SparkSession | None = None, /) -> SparkSession:
    """Validates Spark arg in the 'Dedupe' class"""
    if not spark_session:
        raise ValueError(INVALID_SPARK)
    return spark_session


def validate_keep_arg(keep: Literal["first", "last"]) -> Literal["first", "last"]:
    """Validates Keep arg in the 'Dedupe' class"""
    # TODO: do a type check and TypeError raise here too
    if keep not in ("first", "last"):
        raise ValueError(INVALID_KEEP.format(keep))
    return keep


def validate_strat_arg(strat: BaseStrategy) -> BaseStrategy:
    """Validates that the given 'strategy' is in fact a `BaseStrategy`.

    As used by the strategy manager
    """
    if not isinstance(strat, BaseStrategy):
        raise TypeError(INVALID_STRAT.format(type(strat).__name__))
    return strat


def validate_columns_arg(
    columns: Columns | None,
    is_sequential_applied: bool,
) -> Columns | None:
    """validates inputs to public api 'columns' arg.

    Allowed combinations are:

    - Sequential API: .canonicalize with columns defined
    - Dict API: .canonicalize with no columns defined
    - Rules API: .canonicalize with no columns defined

    Any other combination/repetion raises a value error
    """
    if is_sequential_applied:
        if not columns:
            raise ValueError(INVALID_COLUMNS_EMPTY)

        if isinstance(columns, tuple):
            for label, count in Counter(
                columns,
            ).items():
                if count > 1:
                    raise ValueError(INVALID_COLUMNS_REPEATED.format(label))

    if not is_sequential_applied and columns:
        raise ValueError(INVALID_COLUMNS_NOT_NONE)
    return columns



"""Defines containers for strategies"""

from __future__ import annotations

import warnings
from collections import UserDict
from copy import deepcopy
from typing import Self
from typing import final

from liken._constants import INVALID_DICT_KEY_MSG
from liken._constants import INVALID_DICT_MEMBER_MSG
from liken._constants import INVALID_DICT_VALUE_MSG
from liken._constants import INVALID_FALLBACK_MSG
from liken._constants import INVALID_RULE_EMPTY_MSG
from liken._constants import INVALID_RULE_MEMBER_MSG
from liken._constants import INVALID_SEQUENCE_AFTER_DICT_MSG
from liken._constants import SEQUENTIAL_API_DEFAULT_KEY
from liken._constants import WARN_DICT_REPLACES_SEQUENCE_MSG
from liken._constants import WARN_RULES_REPLACES_RULES_MSG
from liken._strats_library import BaseStrategy
from liken._strats_library import BinaryDedupers
from liken._types import Columns
from liken._validators import validate_strat_arg


# STRATS DICT CONFIG:


@final
class StratsDict(UserDict):
    """Container for combnations of strategies in the Sequential and Dict APIs

    For Sequential API all values (strategies) are added under a default key.

    For Dict API column label(s) (i.e. str or tuple) are the keys."""

    def __setitem__(self, key, value):
        if not isinstance(key, str | tuple):
            raise InvalidStrategyError(INVALID_DICT_KEY_MSG.format(type(key).__name__))
        if not isinstance(value, list | tuple | BaseStrategy):
            raise InvalidStrategyError(INVALID_DICT_VALUE_MSG.format(type(value).__name__))
        if not isinstance(value, BaseStrategy):
            for i, member in enumerate(value):
                if not isinstance(member, BaseStrategy):
                    raise InvalidStrategyError(INVALID_DICT_MEMBER_MSG.format(i, key, type(member).__name__))
        else:
            value = (value,)
        super().__setitem__(key, value)


# STRATS RULES CONFIG


class Rules(tuple):
    """Tuple-like container of strategies.

    Accepts single or multiple strategies where those strategies are passed
    with the ``on`` function.

    Args:
        *strategies: comma separated ``on`` strategies, unpacked.


    Example:
        A single strategy is passed:

            from liken import Dedupe, exact
            from liken.rules import Rules, on

            STRAT = Rules(on("address", exact()))

            lk = Dedupe(df)
            lk.apply(STRAT)

        Multiple strategies are passed:

            from liken import Dedupe, exact
            from liken.rules import Rules, on

            STRAT = Rules(
                on('address', exact()),
                on('email', fuzzy(threshold=0.95)) & on('address', ~isna()),
            )

            lk = Dedupe(df)
            lk.apply(STRAT)
    """

    def __new__(cls, *strategies: On):
        if len(strategies) == 1 and isinstance(strategies[0], tuple):
            strategies = strategies[0]

        if not strategies:
            raise InvalidStrategyError(INVALID_RULE_EMPTY_MSG)

        for i, item in enumerate(strategies):
            if not isinstance(item, On):
                raise InvalidStrategyError(INVALID_RULE_MEMBER_MSG.format(i, type(item).__name__))

        return super().__new__(cls, strategies)


@final
class On:
    """Unit container for a single strategy in the Rules API"""

    def __init__(self, columns: Columns, strat: BaseStrategy):
        self._columns = columns
        self._strat = validate_strat_arg(strat)
        self._strats: list[tuple[Columns, BaseStrategy]] = [(columns, strat)]

    def __and__(self, other: On) -> Self:
        """Overloads `&` operator

        Mutates first instance of On in chain of `&` operates On instances.
        Collectes the combinations of strategies into a single iterable for
        that step. The executor will then apply all these combined strategies
        and select union find components that satisfy all the combinations.

        Usage:
            On("address", exact) & On("email", isna())

        Returns:
            Self, specifically Self of the first On.
        """
        self._strats.append((other._columns, other._strat))
        return self

    @property
    def and_strats(self) -> list[tuple[Columns, BaseStrategy]]:
        """return strategies, sorted such that binary strategies are first.
        This is used for predication.
        """
        return sorted(self._strats, key=lambda x: not isinstance(x[1], BinaryDedupers))

    @property
    def has_any_binary_strat(self) -> bool:
        """whether or not the Rule set of strategies has at least one
        Binary strategy.
        """
        return any([isinstance(x[1], BinaryDedupers) for x in self._strats])

    def __str__(self):
        """string representation

        Parses a single On or combinations of On operated with `&`
        """
        rep = ""
        for cs in self._strats:
            rep += f"on('{cs[0]}', {str(cs[1])}) & "
        return rep[:-3]


# STRATS MANAGER:


@final
class StrategyManager:
    """
    Manage and validate collection(s) of deduplication strategies.

    Supports addition of strategies as part of the three APIs:
    - Sequential
    - Dict
    - Rules

    For Sequential strategies, as instances of `BaseStrategy` are sequentially
    to an idential structure of the Dict API but under a single default
    dictionary key. Keys are columns names, and values are iterables of
    strategies.

    Raises:
        InvalidStrategyError for any misconfigured strategy
    """

    def __init__(self) -> None:
        self._strats: StratsDict | Rules = StratsDict({SEQUENTIAL_API_DEFAULT_KEY: []})
        self.has_applies: bool = False

    @property
    def is_sequential_applied(self) -> bool:
        """checks to see if stratgies are loaded under the default key"""
        return set(self._strats) == {SEQUENTIAL_API_DEFAULT_KEY}

    def apply(self, strat: BaseStrategy | dict | StratsDict | Rules) -> None:
        """Loads a strategy into the manager

        This function currently handles all possible instances of strategy, and
        the implementation achieves this by writing to the strategy dictionary
        or overwriting the dictionary with `Rules`.

        If the input strat is `BaseStrategy` then "Sequential" API is in use. If
        dict (or StratsDict — even though this is not public) then it is the
        "Dict" API. Else "Rules" API is in use.

        Note also that as Rules contains On and combinations of On operated
        with & results in self mutation, need deep copy to allow for
        serialization to Spark workers."""

        # track that at least one apply made
        # if not, used by `Dedupe` to include an exact deduper by default
        self.has_applies = True

        if isinstance(strat, BaseStrategy):
            if not self.is_sequential_applied:
                raise InvalidStrategyError(INVALID_SEQUENCE_AFTER_DICT_MSG)
            self._strats[SEQUENTIAL_API_DEFAULT_KEY].append(strat)
            return

        if isinstance(strat, dict | StratsDict):
            if self._strats[SEQUENTIAL_API_DEFAULT_KEY]:
                warn(WARN_DICT_REPLACES_SEQUENCE_MSG)
            self._strats = StratsDict(strat)
            return

        if isinstance(strat, On):
            strat = (strat,)

        if isinstance(strat, Rules | tuple):
            if isinstance(self._strats, Rules):
                warn(WARN_RULES_REPLACES_RULES_MSG)

            # Contents of Rules is mutable!
            # `On` operated on with `&` results in modified `On`
            # Of which only the first one is preserved
            # To guarantee repeated use of the base class, require deepcopy
            self._strats = Rules(deepcopy(strat))  # type: ignore
            return

        raise InvalidStrategyError(INVALID_FALLBACK_MSG.format(type(strat).__name__))

    def get(self) -> StratsDict | Rules:
        return self._strats

    def pretty_get(self) -> None | str:
        """string representation of strats.

        Output string must be formatted approximately such that it can be used
        with .apply(), i.e. a string representation of one of:
            - BaseStrategy
            - StratsDict
            - Rules
        The seuqneital API with numerous additions of BaseStraegy means there
        is not good way to retried this such that is available to "apply". So,
        default to returning it as a list representation.
        """
        strats = self.get()

        if isinstance(strats, StratsDict):
            # added as BaseStrategy (Sequential API)
            if self.is_sequential_applied:
                # short-circuit; nothing yet applied.
                if not strats[SEQUENTIAL_API_DEFAULT_KEY]:
                    return None

                rep = ""
                for strat in strats[SEQUENTIAL_API_DEFAULT_KEY]:
                    rep += str(strat) + ",\n\t"
                return f"[\n\t{rep[:-3]}\n]"

            # Normal dict API
            rep = ""
            for k, values in strats.items():
                krep = ""
                for v in values:
                    krep += str(v) + ",\n\t"
                rep += f"\n\t'{k}': ({krep[:-3]},),"
            return "{" + rep + "\n}"

        # Rules API
        if isinstance(strats, tuple):
            rep = ""
            for ons in strats:
                rep += str(ons) + ",\n\t"
            return f"Rules(\n\t{rep[:-3]}\n)"

    def reset(self):
        """Reset strategy collection to empty defaultdict"""
        self._strats = StratsDict({SEQUENTIAL_API_DEFAULT_KEY: []})


# EXCEPTIONS:


@final
class InvalidStrategyError(TypeError):
    def __init__(self, msg):
        super().__init__(msg)


# TODO: these warnings come up was "UserWarning". Change.
def warn(msg: str) -> None:
    return warnings.warn(msg, category=UserWarning)


# PUBLIC ON API:


def on(columns: Columns, strat: BaseStrategy, /) -> On:
    """Unit container for a single strategy in the Rules API.

    Operates a "strat" on a "columns". Is provided as comma separated members to
    `Rules`. Allows for "and" chaining via the `&` operator to logically
    compose strategy "rules".

    The `&` ("and") operator is the only supplier logical combination operator
    supplier, as the equivalent to "or" is achieved by comma separating `on`
    calls inside `Rules`. The results of `&` are interepreted as boolean and
    whereby the left-hand deduplication strategy must agree with the right-hand
    strategy for any given pairwise combination.

    Args:
        columns: the label(s) of a column or columns.
        strat: the strategy to apply.

    Returns:
        None

    Example:
        single ``on`` strategy:

            from liken import Dedupe, exact, fuzzy
            from liken.rules import Rules, on, isna, str_endswith

            on("address", exact())

        Strategies combined with ``&``:

            on("email", fuzzy(threshold=0.95)) & on("email", str_endswith("UK"))

        Strategies can be combined with ``&`` for **different** columns:

            on("email", fuzzy(threshold=0.95)) & on("address", ~isna())

        The above can be read as "deduplicate email only when the address field
        is not null":

            >>> df # Before
            +------+-----------+---------------------+
            | id   |  address  |        email        |
            +------+-----------+---------------------+
            |  1   |  london   |  foobar@gmail.com   |
            |  2   |   paris   |  Foobar@gmail.com   |
            |  3   |   null    |  fooBar@gmail.com   |
            +------+-----------+---------------------+

            >>> df # After
            +------+-----------+---------------------+--------------+
            | id   |  address  |        email        | canonical_id |
            +------+-----------+---------------------+--------------+
            |  1   |  london   |  foobar@gmail.com   |       1      |
            |  2   |   paris   |  Foobar@gmail.com   |       1      |
            |  3   |   null    |  fooBar@gmail.com   |       3      |
            +------+-----------+---------------------+--------------+

        Where the first two rows are now linked via the same canonical_id.
    """
    return On(columns, strat)



from liken._strats_library import cosine
from liken._strats_library import exact
from liken._strats_library import fuzzy
from liken._strats_library import jaccard
from liken._strats_library import lsh
from liken._strats_library import tfidf
from liken.dedupe import Dedupe


__all__ = [
    "Dedupe",
    "exact",
    "fuzzy",
    "lsh",
    "tfidf",
    "cosine",
    "jaccard",
]



import random
from typing import Any
from typing import Literal

import pandas as pd
import polars as pl
import pyspark.sql as spark
from faker import Faker
from pyspark.sql import SparkSession


Faker.seed(123)


# CONSTANTS:


# fmt: off


SCHEMA10 = [
    "id", "address", "email", "account",
    "birth_country", "marital_status", "number_children", "property_type", 
    "property_height", "property_area_sq_ft", "property_sea_level_elevation_m", "property_num_rooms"
]

DATA10 = [
    (1, "123ab, OL5 9PL, UK", "bbab@example.com", "reddit", "spain", "married", 1, "rental", None, 545, 5, 3),
    (2, "99 Ambleside avenue park Road, ED3 3RT, Edinburgh, United Kingdom", "awesome_surfer_77@yahoo.com", "reddit", "spain", "married", 1, "rental", None, 452, 6, 3),
    (3, "Calle Ancho, 12, 05688, Rioja, Navarra, Espana", "a@example.com", "facebook", "germany", "single", 2, "rental", 2.5, 623, 5, 3),
    (4, "Calle Sueco, 56, 05688, Rioja, Navarra", "hellothere@example.com", "pinterest", "japan", "married", 0, "owner", 4.0, 2077, 305, 6),
    (5, None, "b@example.com", "linkedin", "france", "married", 1, "rental", 2.7, 1045, 42, 4),
    (6, "C. Ancho 49, 05687, Navarra", "b@example.com", "reddit", "japan", "married", 1, "rental", 2.5, 1323, 132, 4),
    (7, "Ambleside avenue Park Road ED3, UK", "hellthere@example.com", "reddit", "germany", "married", 0, "owner", 2.5, 509, 200, 2),
    (8, "123ab, OL5 9PL, UK", "hellathere@example.com", "facebook", "japan", "single", 3, "owner", 2.5, 500, 300, 3),
    (9, None, "yet.another.email@msn.com", "flickr", "germany", "married", 1, "rental", 2.5, 345, 22, 3),
    (10, "66b Porters street, OL5 9PL, Newark, United Kingdom", "bab@example.com", "flickr", "malaysia", "single", 0, "owner", 2.5, 4000, 25, 8),
]


# fmt: on


SCHEMA10_PLUS = [
    "id",
    "first_name",
    "last_name",
    "email",
    "phone",
    "address",
    "city",
    "country",
    "company",
    "job",
    "date_of_birth",
    "signup_date",
    "salary",
    "is_active",
]


# HELPERS:


fake = Faker()


def _return_df(
    schema: list[str],
    data: list[tuple[Any, ...]],
    backend: Literal["pandas", "polars", "spark"] = "pandas",
    spark_session: SparkSession | None = None,
) -> pd.DataFrame | pl.DataFrame | spark.DataFrame:
    """Returns the dataframe based on the backend"""

    if backend == "pandas":
        return pd.DataFrame(columns=schema, data=data)
    if backend == "polars":
        return pl.DataFrame(schema=schema, data=data, orient="row")
    if backend == "spark":
        if spark_session:
            return spark_session.createDataFrame(schema=schema, data=data)
        raise ValueError("Spark Session not passed yet 'spark' backend requested")
    raise ValueError(f"Expected one of 'pandas', 'polars', 'spark', got '{backend}'")


def maybe_null(value, p):
    return value if random.random() > p else None


def fake_row():
    return (
        fake.uuid4(),
        maybe_null(fake.first_name(), 0.02),
        maybe_null(fake.last_name(), 0.02),
        maybe_null(fake.email(), 0.02),
        maybe_null(fake.phone_number().replace("x", ", "), 0.02),
        maybe_null(fake.address().replace("\n", ", "), 0.02),
        maybe_null(fake.city(), 0.02),
        maybe_null(fake.country(), 0.02),
        maybe_null(fake.company(), 0.02),
        maybe_null(fake.job(), 0.02),
        maybe_null(fake.date_of_birth(minimum_age=18, maximum_age=80), 0.02),
        maybe_null(fake.date_time_this_decade(), 0.02),
        maybe_null(round(random.uniform(30000, 150000), 2), 0.02),
        random.choice([True, False]),
    )


# PUBLIC:


def fake_10(
    backend: Literal["pandas", "polars", "spark"] = "pandas",
    spark_session: SparkSession | None = None,
) -> pd.DataFrame | pl.DataFrame | spark.DataFrame:
    """Synthetic 10 rows.

    Args:
        backend: One of "pandas", "polars" or "spark".
        spark_session: The pyspark spark session if requesting data using
            "spark" backend.

    Returns:
        A dataframe, in the defined backend.

    Raises:
        ValueError: if no spark session passed when requesting a spark dataframe.
    """
    return _return_df(
        schema=SCHEMA10,
        data=DATA10,
        backend=backend,
        spark_session=spark_session,
    )


def fake_1K(
    backend: Literal["pandas", "polars", "spark"] = "pandas",
    spark_session: SparkSession | None = None,
) -> pd.DataFrame | pl.DataFrame | spark.DataFrame:
    """Synthetic 1K (one thousand) rows.

    Args:
        backend: One of "pandas", "polars" or "spark".
        spark_session: The pyspark spark session if requesting data using
            "spark" backend.

    Returns:
        A dataframe, in the defined backend.

    Raises:
        ValueError: if no spark session passed when requesting a spark dataframe.
    """
    data = [fake_row() for _ in range(999)]
    data.append(data[-1])  # duplicate last row for quick-glance

    return _return_df(
        schema=SCHEMA10_PLUS,
        data=data,
        backend=backend,
        spark_session=spark_session,
    )


def fake_100K(
    backend: Literal["pandas", "polars", "spark"] = "pandas",
    spark_session: SparkSession | None = None,
) -> pd.DataFrame | pl.DataFrame | spark.DataFrame:
    """Synthetic 100K (one hundred thousand) rows.

    Args:
        backend: One of "pandas", "polars" or "spark".
        spark_session: The pyspark spark session if requesting data using
            "spark" backend.

    Returns:
        A dataframe, in the defined backend.

    Raises:
        ValueError: if no spark session passed when requesting a spark dataframe.
    """
    data = [fake_row() for _ in range(99_999)]
    data.append(data[-1])  # duplicate last row for quick-glance

    return _return_df(
        schema=SCHEMA10_PLUS,
        data=data,
        backend=backend,
        spark_session=spark_session,
    )


def fake_1M(
    backend: Literal["pandas", "polars", "spark"] = "pandas",
    spark_session: SparkSession | None = None,
) -> pd.DataFrame | pl.DataFrame | spark.DataFrame:
    """Synthetic 1M (one million) rows.

    Args:
        backend: One of "pandas", "polars" or "spark".
        spark_session: The pyspark spark session if requesting data using
            "spark" backend.

    Returns:
        A dataframe, in the defined backend.

    Raises:
        ValueError: if no spark session passed when requesting a spark dataframe.
    """
    data = [fake_row() for _ in range(999_999)]
    data.append(data[-1])  # duplicate last row for quick-glance

    return _return_df(
        schema=SCHEMA10_PLUS,
        data=data,
        backend=backend,
        spark_session=spark_session,
    )



from __future__ import annotations

import os
from typing import Final

import pandas as pd
import polars as pl
import pyspark.sql as spark
from pyspark.sql.types import BooleanType
from pyspark.sql.types import DataType
from pyspark.sql.types import DateType
from pyspark.sql.types import DoubleType
from pyspark.sql.types import FloatType
from pyspark.sql.types import IntegerType
from pyspark.sql.types import LongType
from pyspark.sql.types import StringType
from pyspark.sql.types import TimestampType


# CONSTANTS:

# This must be manually maintained in sync with the `UserDataFrame` type
# E.g. if new backend support is added (from a user point of view)
SUPPORTED_DFS = pd.DataFrame | pl.DataFrame | spark.DataFrame

# Default canonical_id label in the dataframe
CANONICAL_ID: Final[str] = os.environ.get("CANONICAL_ID", "canonical_id")

# Placeholder string for Null values
# This is susceptible to erroneous results e.g. 'str_startswith' is used with `pattern`="n"!
NA_PLACEHOLDER: Final[str] = "na"

# Sequential API use will load to this dictionary key by default:
SEQUENTIAL_API_DEFAULT_KEY: Final[str] = "_default_"


# Pyspark sql conversion types
PYSPARK_TYPES: Final[dict[str, DataType]] = {
    "boolean": BooleanType(),
    "date": DateType(),
    "double": DoubleType(),
    "float": FloatType(),
    "int": IntegerType(),
    "bigint": LongType(),
    "string": StringType(),
    "timestamp": TimestampType(),
}

# ERROR MESSAGES

# For argument validations

INVALID: Final[str] = "Invalid arg: "
INVALID_DF: Final[str] = INVALID + "df must be istance of Pandas, Polars of Spark DataFrames, got {}"
INVALID_SPARK: Final[str] = INVALID + "spark_session must be provided for a spark dataframe"
INVALID_KEEP: Final[str] = INVALID + "keep must be one of 'first' or 'last', got '{}'"
INVALID_STRAT: Final[str] = INVALID + "strat must be instance of BaseStrategy, got {}"
INVALID_COLUMNS_EMPTY: Final[str] = (
    INVALID
    + "columns cannot be None, a column label of tuple of column labels must be provided when using sequential API."
)
INVALID_COLUMNS_REPEATED: Final[str] = INVALID + "columns labels cannot be repeated. Repeated labels: '{}'"
INVALID_COLUMNS_NOT_NONE: Final[str] = (
    INVALID + "columns must be None when using the dict API, as they have been defined as dictionary keys."
)

# strategy collection errors

INVALID_DICT_KEY_MSG: Final[str] = "Invalid type for dict key type: expected str or tuple, got '{}'"
INVALID_DICT_VALUE_MSG: Final[str] = "Invalid type for dict value: expected list, tuple or 'BaseStrategy', got '{}'"
INVALID_DICT_MEMBER_MSG: Final[str] = (
    "Invalid type for dict value member: at index {} for key '{}': 'expected 'BaseStrategy', got '{}'"
)
INVALID_SEQUENCE_AFTER_DICT_MSG: Final[str] = (
    "Cannot apply a 'BaseStrategy' after a strategy mapping (dict) has been set. "
    "Use either individual 'BaseStrategy' instances or a dict of strategies, not both."
)
INVALID_RULE_EMPTY_MSG: Final[str] = "Rules cannot be empty"
INVALID_RULE_MEMBER_MSG: Final[str] = "Invalid Rules element at index {} is not an instance of On, got '{}'"
INVALID_FALLBACK_MSG: Final[str] = "Invalid strategy: Expected a 'BaseStrategy', a dict or 'Rules', got '{}'"

# strategy collection warnings

WARN_DICT_REPLACES_SEQUENCE_MSG: Final[str] = "Replacing previously added sequence strategy with a dict strategy"
WARN_RULES_REPLACES_RULES_MSG: Final[str] = "Replacing previously added 'Rules' strategy with a new 'Rules' strategy"



"""The Rules API defines liken's most flexible and powerful approach.

With this API you can construct rules that combine strategies using and
statements.

Additional boolean choice strategies are defined here — they can be powerfully
combined with the `liken` standard deduplication strategies.
"""

from .._strats_library import isin
from .._strats_library import isna
from .._strats_library import str_contains
from .._strats_library import str_endswith
from .._strats_library import str_len
from .._strats_library import str_startswith
from .._strats_manager import Rules
from .._strats_manager import on


__all__ = [
    "Rules",
    "on",
    "isna",
    "isin",
    "str_startswith",
    "str_contains",
    "str_endswith",
    "str_len",
]



"""Shared liken types"""

from __future__ import annotations

from typing import Literal
from typing import TypeAlias

import pandas as pd
import polars as pl
import pyspark.sql as spark


UserDataFrame: TypeAlias = pd.DataFrame | pl.DataFrame | spark.DataFrame
DataFrameLike: TypeAlias = pd.DataFrame | pl.DataFrame | spark.DataFrame | list[spark.Row]
Columns: TypeAlias = str | tuple[str, ...]  # label(s) that identify attributes of a dataframe for deduplication
Keep: TypeAlias = Literal["first", "last"]  # Canonicalisation rule
SimilarPairIndices: TypeAlias = tuple[int, int]



"""Strategy executors

`SparkExecutor` simply calls a partition processor where each partition will
then be processed with the `LocalExecutor`
"""

# mypy: disable-error-code="no-redef"

from __future__ import annotations

from collections import defaultdict
from collections.abc import Iterator
from functools import partial
from typing import TYPE_CHECKING
from typing import Protocol
from typing import Type
from typing import TypeAlias
from typing import TypeVar
from typing import cast
from typing import final

from networkx.utils.union_find import UnionFind
from pyspark.sql import Row
from pyspark.sql import SparkSession

from liken._constants import CANONICAL_ID
from liken._dataframe import Frame
from liken._dataframe import LocalDF
from liken._dataframe import SparkDF
from liken._strats_library import BaseStrategy
from liken._strats_library import BinaryDedupers
from liken._strats_manager import SEQUENTIAL_API_DEFAULT_KEY
from liken._strats_manager import Rules
from liken._strats_manager import StratsDict
from liken._types import Columns
from liken._types import Keep


if TYPE_CHECKING:
    from liken.dedupe import Dedupe


# TYPES:


SingleComponents: TypeAlias = dict[int, list[int]]
MultiComponents: TypeAlias = dict[tuple[int, ...], list[int]]
F = TypeVar("F", bound=Frame)


# EXECUTORS:


class Executor(Protocol[F]):
    def execute(
        self,
        df: F,
        /,
        *,
        columns: Columns | None,
        strats: StratsDict | Rules,
        keep: Keep,
        drop_duplicates: bool,
        drop_canonical_id: bool,
        id: str | None,
    ) -> F: ...


@final
class LocalExecutor(Executor):
    def execute(
        self,
        df: LocalDF,
        /,
        *,
        columns: Columns | None,
        strats: StratsDict | Rules,
        keep: Keep,
        drop_duplicates: bool,
        drop_canonical_id: bool,
        id: str | None = None,
    ) -> LocalDF:
        """Process a local dataframe according to the strategy collection

        Processing is defined according to whether the collections of
        strategies is:
            - Rules: in which case "and" combinations are allowed
            - StratsDict: in which case handles Sequential and Dict API

        For Rules, predication is implemented if an and combination contains at
        least one Binary Deduper. In that case the binary dedupers are proxy
        WHERE filters that propagate a set of dataframe indice positions to the
        next deduper (most likely a threshold deduper, but optionally binary
        also.)
        """

        del id  # Unused: here for interface symmetry with SparkExecutor

        call_strat = partial(
            self._call_strat,
            drop_duplicates=drop_duplicates,
            keep=keep,
        )

        if isinstance(strats, StratsDict):
            if not columns:
                for col, iter_strats in strats.items():
                    for strat in iter_strats:
                        uf, n = self._build_uf(strat, df, col)
                        components: SingleComponents = self._get_components(uf, n)
                        df = call_strat(strat, components)
            else:
                # For sequential API calls e.g.`.canonicalize("address")`
                for strat in strats[SEQUENTIAL_API_DEFAULT_KEY]:
                    uf, n = self._build_uf(strat, df, columns)
                    components: SingleComponents = self._get_components(uf, n)
                    df = call_strat(strat, components)

        if isinstance(strats, Rules):
            for stage in strats:
                has_any_binary: bool = stage.has_any_binary_strat

                # predication only if at least one binary strategy
                if has_any_binary:
                    indices = set()

                    for col, strat in stage.and_strats:
                        uf, n = self._build_uf(strat, df, col, predicate=indices)

                        components = defaultdict(list)
                        idx: list = sorted(indices)
                        for i in range(n):
                            if not indices:
                                components[uf[i]].append(i)
                            else:
                                components[idx[uf[i]]].append(idx[i])

                        if isinstance(strat, BinaryDedupers):
                            for c in components.values():
                                if len(c) > 1:
                                    indices = indices.union(set(c))

                else:
                    ufs = []

                    for col, strat in stage.and_strats:
                        uf, n = self._build_uf(strat, df, col)
                        ufs.append(uf)
                    components: MultiComponents = self._get_multi_components(ufs, n)

                df = call_strat(strat, components)

        if drop_canonical_id:
            return df.drop_col(CANONICAL_ID)
        return df

    @staticmethod
    def _build_uf(
        strat: BaseStrategy, df: LocalDF, columns: Columns, predicate: set = set()
    ) -> tuple[UnionFind[int], int]:
        return strat.set_frame(df).build_union_find(columns, predicate=predicate)

    @staticmethod
    def _get_components(
        uf: UnionFind[int],
        n: int,
    ) -> SingleComponents:
        components = defaultdict(list)
        for i in range(n):
            components[uf[i]].append(i)
        return components

    @staticmethod
    def _get_multi_components(
        ufs: list[UnionFind[int]],
        n: int,
    ) -> MultiComponents:
        components = defaultdict(list)
        for i in range(n):
            signature = tuple(uf[i] for uf in ufs)
            components[signature].append(i)
        return components

    @staticmethod
    def _call_strat(
        strat: BaseStrategy,
        components: SingleComponents | MultiComponents,
        drop_duplicates: bool,
        keep: Keep,
    ) -> LocalDF:
        return strat.canonicalizer(
            components=components,
            drop_duplicates=drop_duplicates,
            keep=keep,
        )


@final
class SparkExecutor(Executor):
    def __init__(self, spark_session: SparkSession, id: str | None = None):
        self._spark_session = spark_session
        self._id = id

    def execute(
        self,
        df: SparkDF,
        /,
        *,
        columns: Columns | None,
        strats: StratsDict | Rules,
        keep: Keep,
        drop_duplicates: bool,
        drop_canonical_id: bool,
        id: str | None = None,
    ) -> SparkDF:
        """Spark specific deduplication helper

        Maps dataframe partitions to be processed via the RDD API yielding low-
        level list[Rows], which are then post-processed back to a dataframe.

        Args:
            columns: The attribute to deduplicate.
            strats: the collection of strats
        Retuns:
            Instance's _df attribute is updated
        """

        # import in worker node
        from liken.dedupe import Dedupe

        # IMPORTANT: Use local variables, no references to Self
        process_partition = self._process_partition

        rdd = df.mapPartitions(
            lambda partition: process_partition(
                factory=Dedupe,
                partition=partition,
                strats=strats,
                id=id,
                columns=columns,
                drop_duplicates=drop_duplicates,
                keep=keep,
            )
        )

        schema = df._schema

        df = SparkDF(self._spark_session.createDataFrame(rdd, schema=schema), is_init=False)

        if drop_canonical_id:
            return df.drop_col(CANONICAL_ID)
        return df

    @staticmethod
    def _process_partition(
        *,
        factory: Type[Dedupe],
        partition: Iterator[Row],
        strats: StratsDict | Rules,
        id: str | None,
        columns: Columns | None,
        drop_duplicates: bool,
        keep: Keep = "first",
    ) -> Iterator[Row]:
        """process a spark dataframe partition i.e. a list[Row]

        This function is functionality mapped to a worker node. For clean
        separation from the driver, strats are re-instantiated and the main
        liken API is executed *per* worker node.

        Args:
            paritition_iter: a partition
            strats: the collection of strats
            id: the unique identified of the dataset a.k.a "business key"
            columns: the attribute on which to deduplicate

        Returns:
            A list[Row], deduplicated
        """
        # handle empty partitions
        rows: list[Row] = list(partition)
        if not rows:
            return iter([])

        # Core API reused per partition, per worker node
        lk = factory._from_rows(rows)  # type: ignore
        lk.apply(strats)  # type: ignore
        df = lk.canonicalize(
            columns,
            keep=keep,
            drop_duplicates=drop_duplicates,
            id=id,
        )

        return iter(cast(list[Row], df))



"""Synthesised data available for experimentation. Makes data available
according to the chosen backend.
"""

from .synthetic import fake_1K
from .synthetic import fake_1M
from .synthetic import fake_10
from .synthetic import fake_100K


__all__ = [
    "fake_10",
    "fake_1K",
    "fake_100K",
    "fake_1M",
]



"""
This module provides wrappers to allow for a uniform interface across different
backends. The backends covered are:
    - Pandas
    - Polars
    - Spark DataFrames
    - Spark RDDs
    - Spark Rows

Whilst Pandas and Polars wrappers are similarly wrapped, note the following:
- Spark Rows inherits the majority of functionality related to getting
    columns, puting columns, fill na etc
- Conversely, Spark DataFrames take care of adding canonical IDs

Additional Points regarding Spark. Upon initialising the public API with a
Spark DataFrame, the wrapper will call the SparkDF class which will create
canonical IDs. However the output to this is RDDs which are then processed
by the executor into Spark Rows which are dispatched to worker nodes. Spark
Rows can be fully recovered to a Spark DataFrame using the same SparkDF class.

TODO:
    - CanonicalIdMixin should be defined first when inherited
    - A full interface can then be defined
"""

from __future__ import annotations

import warnings
from collections.abc import Hashable
from functools import singledispatch
from typing import Any
from typing import Generic
from typing import Protocol
from typing import Self
from typing import TypeAlias
from typing import TypeVar
from typing import final

import pandas as pd
import polars as pl
import pyarrow as pa
import pyspark.sql as spark
from pyarrow.compute import coalesce
from pyspark.rdd import RDD
from pyspark.sql import Row
from pyspark.sql.types import LongType
from pyspark.sql.types import StructField
from pyspark.sql.types import StructType
from typing_extensions import override

from liken._constants import CANONICAL_ID
from liken._constants import NA_PLACEHOLDER
from liken._constants import PYSPARK_TYPES
from liken._types import Columns
from liken._types import Keep


# TYPES


D = TypeVar("D")  # dataframe
S = TypeVar("S")  # Series


# BASE


class Frame(Generic[D, S]):
    """Base class defining a dataframe wrapper

    Defines inheritable methods as well as some of the interface

    TODO:
        - define a protocol interface
        - tighten generics
    """

    def __init__(self, df: D):
        self._df: D = df

    def unwrap(self) -> D:
        return self._df

    def __getattr__(self, name: str) -> Any:
        """Delegation: use ._df without using property explicitely.

        So, the use of Self even with no attribute returns ._df attribute.
        Therefore calling Self == call Self._df. This is useful as it makes the
        API more concise in other modules.

        For example, as the Dedupe class attribute ._df is an instance of this
        class, it avoids having to do Dedupe()._df._df to access the actual
        dataframe.
        """
        return getattr(self._df, name)

    def _get_col(self, column: str) -> S:
        del column
        raise NotImplementedError

    def _get_cols(self, columns: tuple[str, ...]) -> D:
        del columns
        raise NotImplementedError

    def get_array(self, columns: Columns, with_na: bool = False) -> pa.Array | pa.Table:
        """Generalise the getting of a column, or columns of a df to an array.

        Handles single column and multicolumn. For instances of single column
        the initial column can initially be filled null placeholders, to allow
        for use by strategies. This is optional so that specific strategies
        that do care about nulls are not affected (e.g. IsNA).
        """
        if isinstance(columns, str):
            col: S = self._get_col(columns)
            if with_na:
                return coalesce(col, NA_PLACEHOLDER)
            return col
        return self._get_cols(columns)

    def get_canonical(self) -> pa.Array:
        """Convenience method"""
        return self.get_array(CANONICAL_ID)


# CANONICAL ID


class AddsCanonical(Protocol):
    """Mixin protocol"""

    def _df_as_is(self, df): ...
    def _df_overwrite_id(self, df, id: str): ...
    def _df_copy_id(self, df, id: str): ...
    def _df_autoincrement_id(self, df): ...


class CanonicalIdMixin(AddsCanonical):
    """Defines creation of canonical id upon wrapping a dataframe

    By default a canonical ID is an auto-incrementing numeric field, starting
    from zero.

    However, the canonical ID field can also be:
        - already present in the dataframe as "canonical_id"
        - copied from another "id" field

    In those other instances the resultant canonical id field can therefore
    also be a string field.
    """

    def _add_canonical_id(self, df, id: str | None):

        has_canonical: bool = CANONICAL_ID in df.columns
        id_is_canonical: bool = id == CANONICAL_ID

        if has_canonical:
            if id:
                if id_is_canonical:
                    return self._df_as_is(df)
                # overwrite with id
                return self._df_overwrite_id(df, id)
            warnings.warn(
                f"Canonical ID '{CANONICAL_ID}' already exists. Pass '{CANONICAL_ID}' to `id` arg for consistency",
                category=UserWarning,
            )
            return self._df_as_is(df)
        if id:
            # write new with id
            return self._df_copy_id(df, id)
        # write new auto-incrementing
        return self._df_autoincrement_id(df)


# WRAPPERS


@final
class PandasDF(Frame[pd.DataFrame, pd.Series], CanonicalIdMixin):
    """Pandas DataFrame wrapper"""

    def __init__(self, df: pd.DataFrame, id: str | None = None):
        self._df: pd.DataFrame = self._add_canonical_id(df, id)
        self._id = id

    # CANONICAL ID HELPERS:

    def _df_as_is(self, df: pd.DataFrame) -> pd.DataFrame:
        return df

    def _df_overwrite_id(self, df: pd.DataFrame, id: str) -> pd.DataFrame:
        return df.assign(**{CANONICAL_ID: df[id]})

    def _df_copy_id(self, df: pd.DataFrame, id: str) -> pd.DataFrame:
        return df.assign(**{CANONICAL_ID: df[id]})

    def _df_autoincrement_id(self, df: pd.DataFrame) -> pd.DataFrame:
        return df.assign(**{CANONICAL_ID: pd.RangeIndex(start=0, stop=len(df))})

    # ARROW BACKEND:

    def _get_col(self, column: str) -> pa.Array:
        return pa.array(self._df[column])

    def _get_cols(self, columns: tuple[str, ...]) -> pa.Table:
        return pa.Table.from_pandas(self._df[list(columns)])

    # WRAPPER METHODS:

    def put_col(self, column: str, array: list) -> Self:
        self._df = self._df.assign(**{column: array})
        return self

    def drop_col(self, column: str) -> Self:
        self._df = self._df.drop(columns=column)
        return self

    def drop_duplicates(self, keep: Keep) -> Self:
        self._df = self._df.drop_duplicates(keep=keep, subset=CANONICAL_ID)
        return self


@final
class PolarsDF(Frame[pl.DataFrame, pl.Series], CanonicalIdMixin):
    """Polars DataFrame wrapper"""

    def __init__(self, df: pl.DataFrame, id: str | None = None):
        self._df: pl.DataFrame = self._add_canonical_id(df, id)
        self._id = id

    # CANONICAL ID HELPERS:

    def _df_as_is(self, df: pl.DataFrame) -> pl.DataFrame:
        return df

    def _df_overwrite_id(self, df: pl.DataFrame, id: str) -> pl.DataFrame:
        return df.with_columns(df[id].alias(CANONICAL_ID))

    def _df_copy_id(self, df: pl.DataFrame, id: str) -> pl.DataFrame:
        return df.with_columns(df[id].alias(CANONICAL_ID))

    def _df_autoincrement_id(self, df: pl.DataFrame) -> pl.DataFrame:
        return df.with_columns(pl.arange(0, len(df)).alias(CANONICAL_ID))

    # ARROW BACKEND:

    def _get_col(self, column: str) -> pa.Array:
        return pa.array(self._df.get_column(column))

    def _get_cols(self, columns: tuple[str, ...]) -> pa.Table:
        return self._df.select(columns).to_arrow()

    # WRAPPER METHODS:

    def put_col(self, column: str, array: list) -> Self:
        array = pl.Series(array)  # IMPORTANT; allow list to be assigned to column
        self._df = self._df.with_columns(**{column: array})
        return self

    def drop_col(self, column: str) -> Self:
        self._df = self._df.drop(column)
        return self

    def drop_duplicates(self, keep: Keep) -> Self:
        self._df = self._df.unique(keep=keep, subset=CANONICAL_ID, maintain_order=True)
        return self


SparkObject: TypeAlias = spark.DataFrame | RDD[Row]


@final
class SparkDF(Frame[SparkObject, None], CanonicalIdMixin):
    """Spark DataFrame and Spark RDD wrapper

    This wrapper, contrarily to others does not always add a canonical id. When
    canonical ids are to be added the DataFrame is converted to an RDD for
    downstream processing in Worker nodes.

    The `is_init` flag is then used, when False, to keep a high-level
    DataFrame such that it is easier to drop the canonical_id if within
    `drop_duplicates` regime. Also, then the DataFrame is ready for unwrapping
    and feeding back to the user.

    Note:
        Spark DataFrames have to be converted to RDDs as that is the only way
        to create an autoincrementing field.

    Args:
        df: the dataframe
        id: the label of any other id columns used for creation of canonical_id
        is_init: define whether to route the DataFrame to an RDD along with
            canonical ID creation, or not.
    """

    err_msg = "Method is available for spark Rows only, not spark DataFrame"

    def __init__(
        self,
        df: spark.DataFrame,
        id: str | None = None,
        is_init: bool = True,
    ):
        # new spark plan for safety
        df = df.select("*")

        self._df: SparkObject
        if is_init:
            self._df: RDD[Row] = self._add_canonical_id(df, id)  # type: ignore[no-redef]
        else:
            self._df: spark.DataFrame = df  # type: ignore[no-redef]

        self._id = id

    # CANONICAL ID HELPERS:

    def _df_as_is(self, df: spark.DataFrame) -> RDD[Row]:
        self._schema = df.schema
        return df.rdd

    def _df_overwrite_id(self, df: spark.DataFrame, id: str) -> RDD[Row]:
        df_new: spark.DataFrame = df.drop(CANONICAL_ID)
        self._schema = self._new_schema(df_new, id)
        return df_new.rdd.mapPartitions(
            lambda partition: [Row(**{**row.asDict(), CANONICAL_ID: row[id]}) for row in partition]
        )

    def _df_copy_id(self, df: spark.DataFrame, id: str) -> RDD[Row]:
        self._schema = self._new_schema(df, id)
        return df.rdd.mapPartitions(
            lambda partition: [Row(**{**row.asDict(), CANONICAL_ID: row[id]}) for row in partition]
        )

    def _df_autoincrement_id(self, df: spark.DataFrame) -> RDD[Row]:
        self._schema = self._new_schema(df)
        return df.rdd.zipWithIndex().mapPartitions(
            lambda partition: [Row(**{**row.asDict(), CANONICAL_ID: idx}) for row, idx in partition]
        )

    @staticmethod
    def _new_schema(df: spark.DataFrame, id: str | None = None) -> StructType:
        """Recreate the schema of the dataframe dynamically based on the type
        of the id field.
        """
        fields = df.schema.fields
        if id:
            dtype = dict(df.dtypes)[id]
            id_type = PYSPARK_TYPES[dtype]
        else:
            id_type = LongType()  # auto-incremental is numeric
        fields += [StructField(CANONICAL_ID, id_type, True)]
        return StructType(fields)

    @override
    def unwrap(self) -> spark.DataFrame:
        """Ensure the unwrapped dataframe is always an instance of DataFrame

        Permits the access of the base Dedupe class attribute dataframe to be
        returned as a DataFrame even if no canonicalisation has been applied
        yet. For example this would be needed if inspecting the dataframe as
        contained in an instance of Dedupe having yet to call the canonicalizer
        on the set of strategies"""
        if isinstance(self._df, RDD):
            return self._df.toDF()
        return self._df

    # WRAPPER METHODS:

    def drop_col(self, column: str) -> Self:
        """Only applies to Spark DataFrame to remove canonical ID"""
        if isinstance(self._df, spark.DataFrame):
            self._df = self._df.drop(column)
            return self
        raise NotImplementedError("Cannot drop columns on spark RDD")

    def put_col(self):
        raise NotImplementedError(self.err_msg)

    def _get_col(self):
        raise NotImplementedError(self.err_msg)

    def _get_cols(self):
        raise NotImplementedError(self.err_msg)

    def drop_duplicates(self):
        raise NotImplementedError(self.err_msg)


@final
class SparkRows(Frame[list[spark.Row], list[Any]]):
    """Spark Rows DataFrame

    Spark Rows is what are processed by individual Worker nodes.

    Thus, the `Dedupe` entrypoint is able to process a Spark Rows as `Dedupe`
    will be instantiated in the worker node.
    """

    def __init__(self, df: list[spark.Row]):
        self._df: list[spark.Row] = df

    # ARROW BACKEND:

    def _get_col(self, column: str) -> pa.Array:
        return pa.array([row[column] for row in self._df])

    def _get_cols(self, columns: tuple[str, ...]) -> pa.Table:
        # return [[row[c] for c in columns] for row in self._df]
        data = {col: [row[col] for row in self._df] for col in columns}

        return pa.table(data)

    # WRAPPER METHODS:

    def put_col(self, column: str, array: list) -> Self:
        self._df = [spark.Row(**{**row.asDict(), column: value}) for row, value in zip(self._df, array)]
        return self

    def drop_duplicates(self, keep: Keep) -> Self:

        seen: set[Hashable] = set()
        result: list[Row] = []

        iterable = self._df if keep == "first" else reversed(self._df)

        for row in iterable:
            key = row[CANONICAL_ID]
            if key not in seen:
                seen.add(key)
                result.append(row)

        if keep == "last":
            result.reverse()

        self._df = result
        return self


# DISPATCHER:


@singledispatch
def wrap(
    df: pd.DataFrame | pl.DataFrame | spark.DataFrame | list[spark.Row],
    id: str | None = None,
):
    """
    Wrap the dataframe with instance of `Frame`, for a generic interface
    allowing use of selected methods such as "dropping columns",
    "filling nulls" etc.
    """
    del id  # Unused
    raise NotImplementedError(f"Unsupported data frame: {type(df)}")


@wrap.register(pd.DataFrame)
def _(df, id: str | None = None) -> PandasDF:
    return PandasDF(df, id)


@wrap.register(pl.DataFrame)
def _(df, id: str | None = None) -> PolarsDF:
    return PolarsDF(df, id)


@wrap.register(spark.DataFrame)
def _(df, id: str | None = None) -> SparkDF:
    return SparkDF(df, id)


@wrap.register(list)
def _(df: list[spark.Row], id: str | None) -> SparkRows:
    del id
    return SparkRows(df)


# ACCESSIBLE TYPES


LocalDF: TypeAlias = PandasDF | PolarsDF | SparkRows
DF = TypeVar("DF", SparkDF, LocalDF)



"""Define custom dedupers"""

from collections.abc import Iterable
from collections.abc import Iterator
from functools import wraps
from typing import Callable
from typing import TypeAlias
from typing import final

from typing_extensions import override

from liken._strats_library import ThresholdDedupers
from liken._types import SimilarPairIndices


# TYPES:


PairGenerator: TypeAlias = Callable[[Iterable], Iterable[SimilarPairIndices]]


# CUSTOM:


@final
class Custom(ThresholdDedupers):
    """
    Inherits from Threshold Dedupers for a generalised approach.

    Overrides _gen_similarity_pairs to accept a custom callable, which albeit
    this class being derived from the ThresholdDeduper class, can nevertheless
    be implemented such that it produces binary results.
    """

    def __init__(
        self,
        pair_fn: PairGenerator,
        /,
        *args,
        **kwargs,
    ):
        super().__init__(
            pair_fn=pair_fn,
            *args,
            **kwargs,
        )
        self._pair_fn = pair_fn
        self._args = args
        self._kwargs = kwargs

    @override
    def validate(self, columns):
        """No validation such that custom can be applied to single or
        compound column
        """
        del columns  # Unused
        pass

    @override
    def _gen_similarity_pairs(self, array) -> Iterator[SimilarPairIndices]:
        """generator or function implementation"""
        array: list = array.to_pylist()
        yield from self._pair_fn(array, *self._args, **self._kwargs)

    def __str__(self):
        return self.__repr__()


def register(f: PairGenerator) -> Callable:
    """Register a custom function as a deduplication strategy.

    Custom functions can be registered for use as strategies recognised by the
    ``Dedupe`` class. Use ``register`` as a decorator around the custom callable.

    The custom callable must accept a generic array-like object representing the
    contents of one or more DataFrame columns. The concrete column backing this
    array is resolved only when the strategy is applied.

    The expected function signature is:

        function(array, **kwargs)

    Args:
        f: A custom callable that returns integer pairs of indices identifying
            similar pairs in an array. Accepted callables are functions or
            generators, where generators are preferred.

    Returns:
        Callable

    Raises:
        TypeError: If any positional arguments are used when calling the
            registered strategy.

    Example:
        Registering a custom strategy

            from liken import Dedupe
            from liken.custom import register

            @register
            def custom_deduper(array, **kwargs):
                # your code here
                yield ...

            lk = Dedupe(df)
            lk.apply(custom_deduper(**kwargs))
            df = lk.drop_duplicates("address")

        E.g. the following Custom exact string-length deduplication strategy:

            @register
            def eq_str_len(array):
                n = len(array)
                for i in range(n):
                    for j in range(i + 1, n):
                        if len(array[i]) == len(array[j]):
                            yield i, j

        Applying the strategy:

            lk = Dedupe(df)
            lk.apply(eq_str_len())  # array arg implicitely passed to Dedupe!
            lk.drop_duplicates("address")

        Before:

            +------+-----------+
            | id   | address   |
            +------+-----------+
            |  1   | london    |
            |  2   | paris     |
            |  3   | tokyo     |
            +------+-----------+

        "tokyo" and "paris" have the same length, so reduced:

            +------+-----------+
            | id   | address   |
            +------+-----------+
            |  1   | london    |
            |  2   | paris     |
            +------+-----------+

        Keyword-only enforcement:

            lk.apply(my_func(is_upper_caps=True))  # OK
            lk.apply(my_func(True))                # Raises TypeError
    """

    @wraps(f)
    def wrapper(*args, **kwargs):
        if args:
            raise TypeError(f"{f.__name__} must be called with keyword arguments only")
        return Custom(f, **kwargs)

    return wrapper



"""Defines Deduplication strategies:

Strategies are either:
    - "Threshold" strategies: deduplication is decided according to a
        smiilarity. Routed through main package.
    - "Binary" strategies: deduplication is decided according to discrete
        outcomes. As this choice is fit for combinations using "and"
        operations, this is routed via the "rules" sub-package.
"""

from __future__ import annotations

import re
from collections import defaultdict
from collections.abc import Iterator
from functools import cache
from typing import TYPE_CHECKING
from typing import Iterable
from typing import Protocol
from typing import Self
from typing import final

import numpy as np
import pyarrow as pa
from datasketch import MinHash
from datasketch import MinHashLSH
from networkx.utils.union_find import UnionFind
from numpy.linalg import norm
from rapidfuzz import fuzz
from scipy.sparse import csr_matrix
from sklearn.feature_extraction.text import TfidfVectorizer
from sparse_dot_topn import sp_matmul_topn
from typing_extensions import override

from liken._constants import CANONICAL_ID


if TYPE_CHECKING:
    from liken._dataframe import LocalDF
    from liken._executors import MultiComponents
    from liken._executors import SingleComponents
    from liken._types import Columns
    from liken._types import Keep
    from liken._types import SimilarPairIndices


# INTERFACE:


class Base(Protocol):
    wdf: LocalDF
    with_na_placeholder: bool

    def set_frame(self, wdf: LocalDF) -> Self: ...
    def _gen_similarity_pairs(self, array: pa.Array | pa.Table) -> Iterator[SimilarPairIndices]: ...
    def build_union_find(self, columns: Columns) -> tuple[UnionFind[int], int]: ...
    def canonicalizer(
        self,
        *,
        components: SingleComponents | MultiComponents,
        drop_duplicates: bool,
        keep: Keep,
    ) -> LocalDF: ...
    def str_representation(self, name: str) -> str: ...
    def validate(self, columns: Columns) -> None: ...


# BASE STRATEGY:


class BaseStrategy(Base):
    """
    Base Deduplication class

    By default all dedupers will operate on filled nulls, thus treating them
    as identical instances within a column(s) of values,
    """

    with_na_placeholder: bool = True

    def __init__(self, *args, **kwargs):
        self._init_args = args
        self._init_kwargs = kwargs

    def set_frame(self, wdf: LocalDF) -> Self:
        """Inject dataframe and interface methods"""
        self.wdf: LocalDF = wdf
        return self

    def _gen_similarity_pairs(self, array: pa.Array | pa.Table) -> Iterator[SimilarPairIndices]:
        del array  # Unused
        raise NotImplementedError

    def build_union_find(
        self: Base,
        columns: Columns,
        predicate: set = set(),
    ) -> tuple[UnionFind[int], int]:
        self.validate(columns)

        array: pa.Array | pa.Table = self.wdf.get_array(columns, with_na=self.with_na_placeholder)

        if predicate:
            # subsets the array on predicate indice list
            array: pa.Array | pa.Table = array.take(sorted(predicate))

        n = len(array)

        uf = UnionFind(range(n))
        for i, j in self._gen_similarity_pairs(array):
            uf.union(i, j)

        return uf, n

    def canonicalizer(
        self,
        *,
        components: SingleComponents | MultiComponents,
        drop_duplicates: bool,
        keep: Keep,
    ) -> LocalDF:
        canonicals = self.wdf.get_canonical()

        n = len(canonicals)

        rep_index: dict[int, int] = {}
        for members in components.values():
            if keep == "first":
                rep = min(members)
            elif keep == "last":
                rep = max(members)

            for member in members:
                rep_index[member] = rep

        # for predicated components, default to ith index for non deduplicated rows
        new_canonicals: list = [canonicals[rep_index.get(i, i)].as_py() for i in range(n)]

        self.wdf.put_col(CANONICAL_ID, new_canonicals)

        if not drop_duplicates:
            return self.wdf
        return self.wdf.drop_duplicates(keep=keep)

    def str_representation(self, name: str) -> str:
        args = ", ".join(repr(a) for a in self._init_args)
        kwargs = ", ".join(f"{k}={v!r}" for k, v in self._init_kwargs.items())

        joined = ", ".join(filter(None, [args, kwargs]))
        return f"{name}({joined})"

    def __repr__(self):
        return self.str_representation(self.__class__.__name__)

    def __str__(self):
        # overridable; fall-back to:
        return self.__repr__()


class SingleColumnMixin:
    """
    Validates the column type of deduplication strategy when passed in the
    columns arg. Only single strings allowed.
    """

    def validate(self, columns: Columns) -> None:
        if not isinstance(columns, str):
            raise ValueError("For single column strategies, `columns` must be defined as a string")


class CompoundColumnMixin:
    """
    Validates the column type of deduplication strategy when passed in the
    columns arg. Only tuples of strings allowed
    """

    def validate(self, columns: Columns) -> None:
        if not isinstance(columns, tuple):
            raise ValueError("For compound columns strategies, `columns` must be defined as a tuple")


# EXACT DEDUPER:


@final
class Exact(BaseStrategy):
    """
    Exact deduper.

    Does not accept a validation mixin (and therefore overrides validation)
    As the exact deduper can be applied to single, or compound columns.
    """

    name: str = "exact"

    @override
    def validate(self, columns):
        del columns  # Unused
        pass

    @override
    def _gen_similarity_pairs(self, array: pa.Array | pa.Table):
        buckets = defaultdict(list)

        # single column
        if isinstance(array, pa.Array):
            for i, key in enumerate(array):
                buckets[key].append(i)

        # multi column
        if isinstance(array, pa.Table):
            columns = [array[col] for col in array.column_names]

            n = array.num_rows

            for i in range(n):
                key = tuple(col[i].as_py() for col in columns)
                buckets[key].append(i)

        for indices in buckets.values():
            for i in range(len(indices)):
                for j in range(i + 1, len(indices)):
                    yield indices[i], indices[j]

    def __str__(self):
        return self.str_representation(self.name)


# BINARY DEDUPERS:


class BinaryDedupers(BaseStrategy):
    """
    Defines Binary "choice" deduplications, i.e. those that produce a discrete
    outcome. Any pair of values that satisfies the conditions of a Binary
    Deduper will be deduplicated.

    For example, if StrStartsWith is used for all strings starting with "a",
    then all records for the column starting with the character "a" will
    be canonicalised to the same record.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def _matches(self, value):
        del value  # Unused
        pass

    @override
    def _gen_similarity_pairs(self, array: pa.Array) -> Iterator[SimilarPairIndices]:
        array: list = array.to_pylist()

        n = len(array)
        for i in range(n):
            if not self._matches(array[i]):
                continue
            for j in range(i + 1, n):
                if self._matches(array[j]):
                    yield i, j

    def __invert__(self):
        return _NegatedBinaryDeduper(self)


class _NegatedBinaryDeduper(BinaryDedupers):
    """
    Composable deduplication instance that inverts the results of any binary
    deduper (except IsNA deduper which follows it's own inversion logic).
    """

    def __init__(self, inner: BinaryDedupers):
        self._inner = inner

    def _matches(self, value):
        """simply return the inner classes opposed set of matches"""
        return not self._inner._matches(value)

    def __str__(self):
        return f"~{self._inner}"

    def validate(self, columns):
        "Get the inner instances validation mixin method"
        return getattr(self._inner, "validate")(columns)


@final
class IsNA(
    SingleColumnMixin,
    BinaryDedupers,
):
    """
    Deduplicates all missing / null values into a single group.

    Inversion operator here calls it's own negation class
    """

    name: str = "isna"

    # do NOT want to placehold Null values
    # As we are deduping on them and need to keep them to identify them
    with_na_placeholder: bool = False

    @override
    def _gen_similarity_pairs(self, array: pa.Array):
        array: list = array.to_pylist()

        indices: list[int] = []

        for i, v in enumerate(array):
            if v is None:
                indices.append(i)
                continue

            if v != v:
                indices.append(i)

        for i in range(len(indices)):
            for j in range(i + 1, len(indices)):
                yield indices[i], indices[j]

    def __str__(self):
        return self.str_representation(self.name)

    def __invert__(self):
        return _NotNA()


@final
class _NotNA(
    SingleColumnMixin,
    BinaryDedupers,
):
    """
    Deduplicate all non-NA / non-null values.

    "not a match" for not null does not hold like it does for other Binary
    Dedupers.
    """

    name: str = "~isna"

    with_na_placeholder: bool = False

    @override
    def _gen_similarity_pairs(self, array: pa.Array):
        array: list = array.to_pylist()

        indices: list[int] = []

        for i, v in enumerate(array):
            notna = True
            if v is None:
                notna = False

            elif v != v:
                notna = False

            if notna:
                indices.append(i)

        for i in range(len(indices)):
            for j in range(i + 1, len(indices)):
                yield indices[i], indices[j]

    def __str__(self):
        return self.str_representation(self.name)


@final
class IsIn(
    SingleColumnMixin,
    BinaryDedupers,
):
    """
    Deduplicates all instances of strings that are a member of a defined
    iterable
    """

    name: str = "isin"

    def __init__(self, values: Iterable):
        super().__init__(values=values)
        self._values = values

    @override
    def _matches(self, value: str | None) -> bool:
        return value in self._values

    def __str__(self):
        return self.str_representation(self.name)


@final
class StrLen(
    SingleColumnMixin,
    BinaryDedupers,
):
    """
    Deduplicates all instances of strings that satisfy the bounds in
    (min_len, max_len) where the upper bound can actually be left unbounded.
    """

    name: str = "str_len"

    def __init__(self, min_len: int = 0, max_len: int | None = None):
        super().__init__(min_len=min_len, max_len=max_len)
        self._min_len = min_len
        self._max_len = max_len

    @override
    def _matches(self, value: str | None) -> bool:
        if not value:
            return False
        len_val = len(value)
        if not self._max_len:
            return len_val > self._min_len
        return len_val > self._min_len and len_val <= self._max_len

    def __str__(self):
        return self.str_representation(self.name)


@final
class StrStartsWith(
    SingleColumnMixin,
    BinaryDedupers,
):
    """
    Strings start with canonicalizer.

    Defaults to case sensitive.

    Regex is not supported, please use `StrContains` otherwise.
    """

    name: str = "str_startswith"

    def __init__(self, pattern: str, case: bool = True):
        super().__init__(pattern=pattern, case=case)
        self._pattern = pattern
        self._case = case

    @override
    def _matches(self, value: str | None) -> bool:
        if value is None:
            return False
        return (
            value.startswith(self._pattern)
            #
            if self._case
            else value.lower().startswith(self._pattern.lower())
        )

    def __str__(self):
        return self.str_representation(self.name)


@final
class StrEndsWith(
    SingleColumnMixin,
    BinaryDedupers,
):
    """
    Strings start with canonicalizer.

    Defaults to case sensitive.

    Regex is not supported, please use `StrContains` otherwise.
    """

    name: str = "str_endswith"

    def __init__(self, pattern: str, case: bool = True):
        super().__init__(pattern=pattern, case=case)
        self._pattern = pattern
        self._case = case

    @override
    def _matches(self, value: str | None) -> bool:
        if value is None:
            return False
        return (
            value.endswith(self._pattern)
            #
            if self._case
            else value.lower().endswith(self._pattern.lower())
        )

    def __str__(self):
        return self.str_representation(self.name)


@final
class StrContains(
    SingleColumnMixin,
    BinaryDedupers,
):
    """
    Strings contains canonicalizer.

    Defaults to case sensitive. Supports literal substring or regex search.
    """

    name: str = "str_contains"

    def __init__(self, pattern: str, case: bool = True, regex: bool = False):
        super().__init__(pattern=pattern, case=case, regex=regex)
        self._pattern = pattern
        self._case = case
        self._regex = regex

        if self._regex:
            flags = 0 if self._case else re.IGNORECASE
            self._compiled_pattern = re.compile(self._pattern, flags)

    @override
    def _matches(self, value: str) -> bool:
        if value is None:
            return False

        if self._regex:
            return bool(self._compiled_pattern.search(value))
        else:
            if self._case:
                return self._pattern in value
            else:
                return self._pattern.lower() in value.lower()

    def __str__(self):
        return self.str_representation(self.name)


# THRESHOLD DEDUPERS:


class ThresholdDedupers(BaseStrategy):
    """
    Base instance of dedupers that implement any similarity comparison
    mechanism.
    """

    def __init__(self, threshold: float = 0.95, **kwargs):
        super().__init__(threshold=threshold, **kwargs)
        self._threshold = threshold

        if not (0 <= threshold < 1):
            raise ValueError("The threshold value must be greater or equal to 0 and less than 1")


@final
class Fuzzy(
    SingleColumnMixin,
    ThresholdDedupers,
):
    """
    Fuzzy string matching deduper
    """

    name: str = "fuzzy"

    @staticmethod
    @cache
    def _fuzz_ratio(s1, s2) -> float:
        return fuzz.ratio(s1, s2) / 100

    def _gen_similarity_pairs(self, array: pa.Array) -> Iterator[SimilarPairIndices]:
        array: list = array.to_pylist()

        n = len(array)
        for i in range(n):
            for j in range(i + 1, n):
                if self._fuzz_ratio(array[i], array[j]) > self._threshold:
                    yield i, j

    def __str__(self):
        return self.str_representation(self.name)


@final
class TfIdf(
    SingleColumnMixin,
    ThresholdDedupers,
):
    """
    TF-IDF deduper.

    Additional keywords arguments can be passed to parametrise the vectorizer,
    as listed in the [TF-IDF vectorizer documentation](https://scikit-learn.org/stable/modules/generated/sklearn.feature_extraction.text.TfidfVectorizer.html)
    """

    name: str = "tfidf"

    def __init__(
        self,
        ngram: int | tuple[int, int] = 3,
        threshold: float = 0.95,
        topn: int = 2,
        **kwargs,
    ):
        super().__init__(
            threshold=threshold,
            ngram=ngram,
            topn=topn,
            **kwargs,
        )
        self._ngram = ngram
        self._threshold = threshold
        self._topn = topn
        self._kwargs = kwargs

    def _vectorize(self) -> TfidfVectorizer:
        ngram_range = (self._ngram, self._ngram) if isinstance(self._ngram, int) else self._ngram

        return TfidfVectorizer(
            analyzer="char",
            ngram_range=ngram_range,
            **self._kwargs,
        )

    def _get_sparse_matrix(self, array: list) -> csr_matrix:
        """sparse matrix of similarities, given the top N best matches"""

        vectorizer = self._vectorize()
        matrix = vectorizer.fit_transform(array)
        return sp_matmul_topn(
            matrix,
            matrix.T,
            top_n=self._topn,
            threshold=self._threshold,
            sort=True,
        )

    def _gen_similarity_pairs(self, array: pa.Array) -> Iterator[SimilarPairIndices]:
        """Extract arrays based on similarity scores

        Filter's out _approximate_ perfect scores (i.e. decimal handling) and
        loads up results into a tuple of arrays"""
        array: list = array.to_pylist()

        sparse = self._get_sparse_matrix(array)

        sparse_coo = sparse.tocoo()

        rows, cols = sparse_coo.row, sparse_coo.col

        for i in range(len(rows)):
            yield rows[i], cols[i]

    def __str__(self):
        return self.str_representation(self.name)


@final
class LSH(
    SingleColumnMixin,
    ThresholdDedupers,
):
    """
    Locality Sensitive Hashing deduper
    """

    name: str = "lsh"

    def __init__(
        self,
        ngram: int = 3,
        num_perm: int = 128,
        threshold: float = 0.95,
    ):
        super().__init__(
            threshold=threshold,
            ngram=ngram,
            num_perm=num_perm,
        )
        self._ngram = ngram
        self._threshold = threshold
        self._num_perm = num_perm

    def _gen_token(self, text: pa.Scalar) -> Iterator:
        for i in range(len(text) - self._ngram + 1):
            yield text[i : i + self._ngram]

    def _build_minhashes(self, array: list) -> list[MinHash]:
        minhashes: list[MinHash] = []
        for value in array:
            m = MinHash(num_perm=self._num_perm)
            for token in self._gen_token(value):
                m.update(token.encode("utf8"))
            minhashes.append(m)
        return minhashes

    def _lsh(self, minhashes: list[MinHash]) -> MinHashLSH:
        lsh = MinHashLSH(
            threshold=self._threshold,
            num_perm=self._num_perm,
        )

        for i, m in enumerate(minhashes):
            lsh.insert(i, m)

        return lsh

    def _gen_similarity_pairs(self, array: pa.Array) -> Iterator[SimilarPairIndices]:
        array: list = array.to_pylist()

        minhashes: list[MinHash] = self._build_minhashes(array)
        lsh: MinHashLSH = self._lsh(minhashes)

        for idx, minhash in enumerate(minhashes):
            for idy in lsh.query(minhash):
                if idx < idy:
                    yield idx, idy

    def __str__(self):
        return self.str_representation(self.name)


# COMPOUND COLUMN:


@final
class Jaccard(
    CompoundColumnMixin,
    ThresholdDedupers,
):
    """
    Deduplicate sets where such sets contain categorical data.
    """

    name: str = "jaccard"

    def _gen_similarity_pairs(self, array: pa.Table) -> Iterator[SimilarPairIndices]:
        columns = [array[col] for col in array.column_names]
        n = array.num_rows

        # Build row sets directly
        sets = [{col[i].as_py() for col in columns if col[i].as_py() is not None} for i in range(n)]

        for idx in range(n):
            for idy in range(idx + 1, n):
                intersection = sets[idx] & sets[idy]

                if not intersection:
                    continue  # no match

                union = sets[idx] | sets[idy]

                if not union:
                    continue  # zero div: guardrail

                if len(intersection) / len(union) > self._threshold:
                    yield idx, idy

    def __str__(self):
        return self.str_representation(self.name)


@final
class Cosine(
    CompoundColumnMixin,
    ThresholdDedupers,
):
    """
    Deduplicate sets where such sets contain numeric data.
    """

    name: str = "cosine"

    def _gen_similarity_pairs(self, array: pa.Table) -> Iterator[SimilarPairIndices]:

        # TODO: See Jaccard implementation for consideration in not using numpy here.
        # For the meantime OK, as no portability issues, plus we get vectorization.
        columns: list = [array[col].to_numpy(zero_copy_only=False) for col in array.column_names]

        matrix = np.column_stack(columns)

        n = len(matrix)

        for idx in range(n):
            for idy in range(idx + 1, n):
                arrx = matrix[idx]
                arry = matrix[idy]

                mask = ~np.isnan(arrx) & ~np.isnan(arry)

                arrx_masked = arrx[mask]
                arry_masked = arry[mask]
                product = np.dot(arrx_masked, arry_masked)

                if not product:
                    continue  # no match

                norms = norm(arrx_masked) * norm(arry_masked)

                if not norms:
                    continue  # zero div: guardrail

                if product / norms > self._threshold:
                    yield idx, idy

    def __str__(self):
        return self.str_representation(self.name)


# PUBLIC PKG:


def exact() -> BaseStrategy:
    """Exact Deduplication.

    Can deduplicate a single column, or multiple columns.

    If no strategies are applied to `Dedupe`, `exact` is applied by default.

    Returns:
        Instance of `BaseStrategy`..

    Example:
        Applied to a single column:

            from liken import Dedupe, exact

            lk = Dedupe(df)
            lk.apply(exact())
            df = lk.drop_duplicates("address")

        Applied to multiple columns:

            lk = Dedupe(df)
            lk.apply(exact())
            df = lk.drop_duplicates(("address", "email"))

        E.g.

            >>> df # Before
            +------+-----------+--------------------+
            | id   |  address  |        email       |
            +------+-----------+--------------------+
            |  1   |  london   |  fizzpop@gmail.com |
            |  2   |   null    |  foobar@gmail.com  |
            |  3   |   null    |  foobar@gmail.com  |
            +------+-----------+--------------------+

            >>> df # After
            +------+-----------+---------------------+
            | id   |  address  |        email        |
            +------+-----------+---------------------+
            |  1   |  london   |  fizzpop@gmail.com  |
            |  2   |   null    |  foobar@gmail.com   |
            +------+-----------+---------------------+

        By default `exact` is used when no stratgies are explicitely applied:

            lk = Dedupe(df)
            lk.drop_duplicates("address")   # OK, still dedupes.
    """
    return Exact()


def fuzzy(threshold: float = 0.95) -> BaseStrategy:
    """Near string deduplication.

    Usage is on single columns of a dataframe.

    Args:
        threshold: the minimum threshold at which similarity between two pairs
            of values will be considered valid for deduplication.

    Returns:
        Instance of `BaseStrategy`.

    Example:
        Applied to a single column:

            from liken import Dedupe, fuzzy

            lk = Dedupe(df)
            lk.apply({"address": fuzzy(threshold=0.8)})
            df = lk.drop_duplicates(keep="last")

        E.g.

            >>> df # Before
            +------+-----------+----------------------+
            | id   |  address  |         email        |
            +------+-----------+----------------------+
            |  1   |  london   |  fizzpop@gmail.com   |
            |  2   |   null    |  foobar@gmail.com    |
            |  3   |  london   |  foobar@gmail.co.uk  |
            +------+-----------+----------------------+

            >>> df # After
            +------+-----------+----------------------+
            | id   |  address  |         email        |
            +------+-----------+----------------------+
            |  1   |  london   |  fizzpop@gmail.com   |
            |  3   |  london   |  foobar@gmail.co.uk  |
            +------+-----------+----------------------+
    """
    return Fuzzy(threshold=threshold)


def tfidf(
    threshold: float = 0.95,
    ngram: int | tuple[int, int] = 3,
    topn: int = 2,
    **kwargs,
) -> BaseStrategy:
    """Near string deduplication using term frequency, inverse document
    frequency.

    Usage is on single columns of a dataframe. `tfidf` is a tuneable deduper.
    Experimentation is required for optimal use.

    Args:
        threshold: the minimum threshold at which similarity between two pairs
            of values will be considered valid for deduplication.
        ngram: the number of character ngrams to consider. For the `tfidf`
            implementation this is the ngram bounded range. If you pass this as
            an integer you are saying the bounds are the same. E.g. `ngram=1`
            is equivalent to the range bounded over (1, 1) (i.e. unigrams
            only). `ngram=(1, 2)` is unigrams and bigrams. Increasing ngrams
            reduces overall deduplication. However, too small an `ngram` may
            result in false positives.
        topn: the number of best matches to consider when building similarity
            matrices.
        **kwargs: additional kwargs as accepted in sklearn's [Tfidf
            Vectorizer](https://scikit-learn.org/stable/modules/generated/sklearn.feature_extraction.text.TfidfVectorizer.html)

    Returns:
        Instance of `BaseStrategy`.

    Example:
        Applied to a single column:

            from liken import Dedupe, tfidf

            lk = Dedupe(df)
            lk.apply({"address": tfidf(threshold=0.8, ngram=1)})
            df = lk.drop_duplicates(keep="last")

        E.g.

            >>> df # Before
            +------+-----------+----------------------+
            | id   |  address  |         email        |
            +------+-----------+----------------------+
            |  1   |  london   |  fizzpop@gmail.com   |
            |  2   |   null    |  foobar@gmail.com    |
            |  3   |  london   |  foobar@gmail.co.uk  |
            +------+-----------+----------------------+

            >>> df # After
            +------+-----------+----------------------+
            | id   |  address  |         email        |
            +------+-----------+----------------------+
            |  1   |  london   |  fizzpop@gmail.com   |
            |  3   |  london   |  foobar@gmail.co.uk  |
            +------+-----------+----------------------+

        Note that the same deduper with `ngram=2` does not deduplicate any
        records in the above example.
    """
    return TfIdf(threshold=threshold, ngram=ngram, topn=topn, **kwargs)


def lsh(
    threshold: float = 0.95,
    ngram: int = 3,
    num_perm: int = 128,
) -> BaseStrategy:
    """Near string deduplication using locality sensitive hashing (LSH).

    Usage is on single columns of a dataframe. `lsh` is a tuneable deduper.
    Experimentation is required for optimal use.

    Args:
        threshold: the minimum threshold at which similarity between two pairs
            of values will be considered valid for deduplication.
        ngram: the number of character ngrams to consider. For `lsh`, and
            unlike the `tfidf` implementation, this is single integer ngram
            number. So, `ngram=1` is only unigrams. Increasing ngrams reduces
            overall deduplication. However, too small an `ngram` may result in false
            positives.
        num_perm: the number of MinHash permutations used to approximate
            similarity. Increasing this generally produces better matches, at
            greater computational cost. Very low numbers of permutations (< 64)
            can produce unreliable results.

    Returns:
        Instance of `BaseStrategy`.

    Example:
        Applied to a single column:

            from liken import Dedupe, lsh

            lk = Dedupe(df)
            lk.apply({"address": lsh(threshold=0.8, ngram=1)})
            df = lk.drop_duplicates(keep="last")

        E.g.

            >>> df
            +------+-----------+----------------------+
            | id   |  address  |         email        |
            +------+-----------+----------------------+
            |  1   |  london   |  fizzpop@gmail.com   |
            |  2   |   null    |  foobar@gmail.com    |
            |  3   |  london   |  foobar@gmail.co.uk  |
            +------+-----------+----------------------+

            >>> df # After deduplication
            +------+-----------+----------------------+
            | id   |  address  |         email        |
            +------+-----------+----------------------+
            |  1   |  london   |  fizzpop@gmail.com   |
            |  3   |  london   |  foobar@gmail.co.uk  |
            +------+-----------+----------------------+
    """
    return LSH(threshold=threshold, ngram=ngram, num_perm=num_perm)


def jaccard(threshold: float = 0.95) -> BaseStrategy:
    """Multi-column deduplication using jaccard similarity.

    Usage is on multiple columns of a dataframe. Appropriate for categorical
    data. Null types are handled out-of-box with jaccard, they are simply
    considered another category of a given field.

    Args:
        threshold: the minimum threshold at which similarity between two pairs
            of values will be considered valid for deduplication.

    Returns:
        Instance of `BaseStrategy`.

    Example:
        Applied to multiple columns:

            from liken import Dedupe, jaccard

            lk = Dedupe(df)
            lk.apply(jaccard())
            df = lk.drop_duplicates(
                ("account", "status", "country", "property"),
                keep="first",
            )

        E.g.

            >>> df
            +------+-----------+----------+----------+-----------+
            | id   |  account  |  status  |  country |  property |
            +------+-----------+----------+----------+-----------+
            |  1   |  reddit   |  married |    UK    |  house    |
            |  2   |  flickr   |  married |    UK    |  house    |
            |  3   | pinterest |  single  |  Germany |  flat     |
            +------+-----------+----------+----------+-----------+

            >>> df # After deduplication
            +------+-----------+----------+----------+-----------+
            | id   |  account  |  status  |  country |  property |
            +------+-----------+----------+----------+-----------+
            |  1   |  reddit   |  married |    UK    |  house    |
            |  3   | pinterest |  single  |  Germany |  flat     |
            +------+-----------+----------+----------+-----------+
    """
    return Jaccard(threshold=threshold)


def cosine(threshold: float = 0.95) -> BaseStrategy:
    """Multi-column deduplication using cosine similarity.

    Usage is on multiple columns of a dataframe. Appropriate for numerical
    data.

    Args:
        threshold: the minimum threshold at which similarity between two pairs
            of values will be considered valid for deduplication.

    Returns:
        Instance of `BaseStrategy`.

    Note:
        In the case of null types, that column is ignore, and only the
        similarity is taken of the remaining columns is taken.

        So, if deduplicating columns `col_1`, `col_2` and `col_3` with `cosine`,
        any similarity is usually the dot product for a given pairwise evaluation
        i.e.

            (`col_1i`, `col_2i`, `col_3i`) . (`col_1j`, `col_2j`, `col_3j`)

        However, if `col_1i` is Null then the following is evaluated:

            (`col_2i`, `col_3i`) . (`col_2j`, `col_3j`)

        Additionally, if `col_j2` is *also* Null then the following is evaluated:

            (`col_3i`) . (`col_3j`)

        Taking this into account you may find it best to avoid cosine similarity
        calculations for sparse datasets. Alternatively, you may opt to your
        approach by either preprocessing the Nulls beforehand, or, by
        limiting yourself to using the `cosine` deduplicator with the `Rules`
        API using combinations for non null fields, e.g.

            STRAT = Rules(
                on(
                    ("col_1", "col_2", "col_3"),
                    cosine(),
                )
                #
                & on("col_1", ~isna()),
            )

    Warning:
        Normalization is a standard approach to ensure that the results of
        cosine similarity are valid. Consider [standard
        approaches](https://scikit-learn.org/stable/modules/preprocessing.html#normalization)

    Example:
        Applied to multiple columns:

            from liken import Dedupe, cosine

            lk = Dedupe(df)
            lk.apply(cosine())
            df = lk.drop_duplicates(
                ("surface are", "ceiling height", "building age", "num_rooms"),
                keep="first",
            )
    """
    return Cosine(threshold=threshold)


# RULES SUB PKG


def isna() -> BaseStrategy:
    """Discrete deduper on null/None values.

    Usage is on a single column of a dataframe. Available as the inversion, i.e.
    "not null" using inversion operator: `~isna()`.

    Returns:
        Instance of `BaseStrategy`.

    Example:
        Applied to a single column:

            from liken import Dedupe, exact
            from liken.rules import Rules, on, isna

            STRAT = Rules(on("email", exact()) & on("address", ~isna()))

            lk = Dedupe(df)
            lk.apply(STRAT)
            df = lk.drop_duplicates(keep="last")

            >>> df # before
            +------+-----------+---------------------+
            | id   |  address  |         email       |
            +------+-----------+---------------------+
            |  1   |  london   |  fizzpop@yahoo.com  |
            |  2   |  london   |  fizzpop@yahoo.com  |
            |  3   |   null    |  foobar@gmail.com   |
            |  4   |   null    |  foobar@gmail.com   |
            +------+-----------+---------------------+

            >>> df # after
            +------+-----------+---------------------+
            | id   |  address  |         email       |
            +------+-----------+---------------------+
            |  2   |  london   |  fizzpop@yahoo.com  |
            |  3   |   null    |  foobar@gmail.com   |
            |  4   |   null    |  foobar@gmail.com   | # Not deduped!
            +------+-----------+---------------------+
    """
    return IsNA()


def isin(values: Iterable) -> BaseStrategy:
    """Discrete deduper for membership testing.

    Usage is on a single column of a dataframe. Available as the inversion, i.e.
    "not in" using inversion operator: `~isin()`.

    Returns:
        Instance of `BaseStrategy`.

    Example:
        Applied to a single column:

            from liken import Dedupe, exact
            from liken.rules import Rules, on, isin

            STRAT = Rules(on("address", isin(values="london")))

            lk = Dedupe(df)
            lk.apply(STRAT)
            df = lk.drop_duplicates(keep="last")

            >>> df # before
            +------+-----------+---------------------+
            | id   |  address  |         email       |
            +------+-----------+---------------------+
            |  1   |  london   |  fizzpop@yahoo.com  |
            |  2   |  london   |   hello@yahoo.com   |
            |  3   |   null    |  foobar@gmail.com   |
            |  4   |   null    |  random@gmail.com   |
            |  5   |  london   |  butterfly@msn.jp   |
            +------+-----------+---------------------+

            >>> df # after
            +------+-----------+---------------------+
            | id   |  address  |         email       |
            +------+-----------+---------------------+
            |  3   |   null    |  foobar@gmail.com   |
            |  4   |   null    |  random@gmail.com   |
            |  5   |  london   |  butterfly@msn.jp   |
            +------+-----------+---------------------+
    """
    return IsIn(values=values)


def str_len(min_len: int = 0, max_len: int | None = None) -> BaseStrategy:
    """Discrete deduper on string length.

    Usage is on a single column of a dataframe. Available as the inversion, i.e.
    "not the defined length" using inversion operator: `~str_len()`.

    Deduplication will happen over the bounded lengths defined by `min_len` and
    `max_len`. The upper end of the range can be left unbounded. For
    deduplication over an exact length use `max_len = min_len + 1`.

    Args:
        min_len: the lower bound of lengths considered
        max_len: the upper bound of lengths considered. Can be left unbounded.

    Returns:
        Instance of `BaseStrategy`.

    Example:
        Applied to a single column:

            from liken import Dedupe, exact
            from liken.rules import Rules, on, isna

            STRAT = Rules(on("email", exact()) & on("email", str_len(min_len=10)))

            lk = Dedupe(df)
            lk.apply(STRAT)
            df = lk.drop_duplicates(keep="last")

            >>> df # before
            +------+-----------+---------------------+
            | id   |  address  |         email       |
            +------+-----------+---------------------+
            |  1   |  london   |  fizzpop@yahoo.com  |
            |  2   |   tokyo   |  fizzpop@yahoo.com  |
            |  3   |   paris   |       a@msn.fr      |
            |  4   |   nice    |       a@msn.fr      |
            +------+-----------+---------------------+

            >>> df # after
            +------+-----------+---------------------+
            | id   |  address  |         email       |
            +------+-----------+---------------------+
            |  2   |   tokyo   |  fizzpop@yahoo.com  |
            |  3   |   paris   |       a@msn.fr      |
            |  4   |   nice    |       a@msn.fr      |
            +------+-----------+---------------------+
    """
    return StrLen(min_len=min_len, max_len=max_len)


def str_startswith(pattern: str, case: bool = True) -> BaseStrategy:
    """Discrete deduper on strings starting with a pattern.

    Usage is on a single column of a dataframe. Available as the inversion, i.e.
    "not starting with pattern" using inversion operator: `~str_startswith()`.

    Deduplication will happen for any pairwise matches that have the same
    `pattern`. Case sensitive unless optionally removed.

    Args:
        pattern: the pattern that the string starts with to be deduplicated
        case: case sensitive, or not.

    Returns:
        Instance of `BaseStrategy`.

    Example:
        Applied to a single column:

            from liken import Dedupe, exact
            from liken.rules import Rules, on, str_startswith

            STRAT = Rules(
                on("email", exact())
                & on(
                    "email",
                    str_startswith(pattern="f", case=True),
                )
            )

            lk = Dedupe(df)
            lk.apply(STRAT)
            df = lk.drop_duplicates(keep="first")

            >>> df
            +------+-----------+---------------------+
            | id   |  address  |         email       |
            +------+-----------+---------------------+
            |  1   | new york  |  fizzpop@yahoo.com  |
            |  2   |   london  |  foobar@gmail.co.uk |
            |  3   | marseille |   Flipflop@msn.fr   |
            |  4   |  chicago  |    random@aol.com   |
            +------+-----------+---------------------+

            >>> df # after
            +------+-----------+---------------------+
            | id   |  address  |         email       |
            +------+-----------+---------------------+
            |  1   | new york  |  fizzpop@yahoo.com  |
            |  3   | marseille |   Flipflop@msn.fr   |
            |  4   |  chicago  |    random@aol.com   |
            +------+-----------+---------------------+
    """
    return StrStartsWith(pattern=pattern, case=case)


def str_endswith(pattern: str, case: bool = True) -> BaseStrategy:
    """Discrete deduper on strings ending with a pattern.

    Usage is on a single column of a dataframe. Available as the inversion, i.e.
    "not ending with pattern" using inversion operator: `~str_endswith()`.

    Deduplication will happen for any pairwise matches that have the same
    `pattern`. Case sensitive unless optionally removed.

    Args:
        pattern: the pattern that the string ends with to be deduplicated
        case: case sensitive, or not.

    Returns:
        Instance of `BaseStrategy`.

    Example:
        Applied to a single column:

            from liken import Dedupe, exact
            from liken.rules import Rules, on, str_endswith

            STRAT = Rules(
                on("email", exact())
                & on(
                    "email",
                    str_endswith(pattern=".com", case=False),
                )
            )

            lk = Dedupe(df)
            lk.apply(STRAT)
            df = lk.drop_duplicates(keep="first")

            >>> df
            +------+-----------+---------------------+
            | id   |  address  |         email       |
            +------+-----------+---------------------+
            |  1   | new york  |  fizzpop@yahoo.Com  |
            |  2   |   london  |  foobar@gmail.co.uk |
            |  3   | marseille |   Flipflop@msn.fr   |
            |  4   |  chicago  |    random@aol.com   |
            +------+-----------+---------------------+

            >>> df # after
            +------+-----------+---------------------+
            | id   |  address  |         email       |
            +------+-----------+---------------------+
            |  1   | new york  |  fizzpop@yahoo.Com  |
            |  2   |   london  |  foobar@gmail.co.uk |
            |  3   | marseille |   Flipflop@msn.fr   |
            +------+-----------+---------------------+
    """
    return StrEndsWith(pattern=pattern, case=case)


def str_contains(
    pattern: str,
    case: bool = True,
    regex: bool = False,
) -> BaseStrategy:
    """Discrete deduper on general string patterns with regex.

    Usage is on a single column of a dataframe. Available as the inversion, i.e.
    "not containing pattern" using inversion operator: `~str_contains()`.

    Deduplication will happen for any pairwise matches that have the same
    `pattern`. Case sensitive unless optionally removed. Pattern can include
    regex patterns if passed with `regex` arg.

    Args:
        pattern: the pattern that the string ends with to be deduplicated
        case: case sensitive, or not.
        regex: uses regex patterns, or not.

    Returns:
        Instance of `BaseStrategy`.

    Example:
        Applied to a single column:

            from liken import Dedupe, exact
            from liken.rules import Rules, on, str_contains

            STRAT = Rules(
                on("email", exact())
                & on(
                    "email",
                    str_contains(pattern=r"05\d{3}", regex=True),
                )
            )

            lk = Dedupe(df)
            lk.apply(STRAT)
            df = lk.canonicalize(keep="first")

            >>> df
            +------+-----------------------------+
            | id   |           address           |
            +------+-----------------------------+
            |  1   | 12 calle girona, 05891, ES  |
            |  2   |  1A avenida palmas, 05562   |
            |  3   |      901, Spain, 05435      |
            |  4   |     12, santiago, 09945     |
            +------+-----------------------------+

            >>> df # after
            +------+-----------------------------+---------------+
            | id   |           address           |  canonical_id |
            +------+-----------------------------+---------------+
            |  1   | 12 calle girona, 05891, ES  |        1      |
            |  2   |  1A avenida palmas, 05562   |        1      |
            |  3   |      901, Spain, 05435      |        1      |
            |  4   |     12, santiago, 09945     |        4      |
            +------+-----------------------------+---------------+
    """
    return StrContains(pattern=pattern, case=case, regex=regex)
